from __future__ import annotations
from qtpy import QtWidgets as QtW, QtCore
from qtpy.QtCore import Qt
import numpy as np

from ._table_base import QBaseTable, QMutableTable


class TableItemDelegate(QtW.QStyledItemDelegate):
    """Displays table widget items with properly formatted numbers."""

    def __init__(self, parent: QtCore.QObject | None = None, ndigits: int = 4) -> None:
        super().__init__(parent)
        self.ndigits = ndigits

    def displayText(self, value, locale):
        return super().displayText(self._format_number(value), locale)

    def createEditor(
        self, parent: QtW.QWidget, option, index: QtCore.QModelIndex
    ) -> QtW.QWidget:
        """Create different type of editors for different dtypes."""
        qtable: QBaseTable = parent.parent()
        table = qtable.parent()
        if isinstance(table, QMutableTable):
            df = table.model().df
            row = index.row()
            col = index.column()
            if row >= df.shape[0] or col >= df.shape[1]:
                # out of bounds
                return super().createEditor(parent, option, index)

            dtype: np.dtype = df.dtypes.values[col]
            if dtype == "category":
                # use combobox for categorical data
                cbox = QtW.QComboBox(parent)
                choices = list(map(str, dtype.categories))
                cbox.addItems(choices)
                cbox.setCurrentIndex(choices.index(df.iat[row, col]))
                cbox.currentIndexChanged.connect(qtable.setFocus)
                return cbox
            elif dtype == "bool":
                # use checkbox for boolean data
                cbox = QtW.QComboBox(parent)
                choices = ["True", "False"]
                cbox.addItems(choices)
                cbox.setCurrentIndex(0 if df.iat[row, col] else 1)
                cbox.currentIndexChanged.connect(qtable.setFocus)
                return cbox
            elif dtype.kind == "M":
                dt = QtW.QDateTimeEdit(parent)
                dt.setDateTime(df.iat[row, col].to_pydatetime())
                return dt

        return super().createEditor(parent, option, index)

    def setEditorData(self, editor: QtW.QWidget, index: QtCore.QModelIndex) -> None:
        super().setEditorData(editor, index)
        if isinstance(editor, QtW.QComboBox):
            editor.showPopup()
        return None

    def setModelData(
        self,
        editor: QtW.QWidget,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ) -> None:
        if isinstance(editor, QtW.QDateTimeEdit):
            dt = editor.dateTime().toPyDateTime()
            model.setData(index, dt, Qt.ItemDataRole.EditRole)
        else:
            return super().setModelData(editor, model, index)

    # modified from magicgui
    def _format_number(self, text: str) -> str:
        """convert string to int or float if possible"""
        try:
            value: int | float | None = int(text)
        except ValueError:
            try:
                value = float(text)
            except ValueError:
                value = None

        ndigits = self.ndigits

        if isinstance(value, (int, float)):
            if 0.1 <= abs(value) < 10 ** (ndigits + 1) or value == 0:
                text = str(value) if isinstance(value, int) else f"{value:.{ndigits}f}"
            else:
                text = f"{value:.{ndigits-1}e}"

        return text



from __future__ import annotations
from typing import Callable
from qtpy.QtWidgets import QMessageBox, QAction
from qtpy import QtWidgets as QtW, QtGui
from qtpy.QtCore import Qt
from enum import Enum
from functools import reduce
from operator import or_


class MessageBoxMode(Enum):
    ERROR = "error"
    WARNING = "warn"
    INFO = "info"
    QUESTION = "question"
    ABOUT = "about"


_QMESSAGE_MODES = {
    MessageBoxMode.ERROR: QMessageBox.critical,
    MessageBoxMode.WARNING: QMessageBox.warning,
    MessageBoxMode.INFO: QMessageBox.information,
    MessageBoxMode.QUESTION: QMessageBox.question,
    MessageBoxMode.ABOUT: QMessageBox.about,
}


def show_messagebox(
    mode: str | MessageBoxMode = MessageBoxMode.INFO,
    title: str = None,
    text: str = None,
    parent=None,
) -> bool:
    """
    Freeze the GUI and open a messagebox dialog.

    Parameters
    ----------
    mode : str or MessageBoxMode, default is MessageBoxMode.INFO
        Mode of message box. Must be "error", "warn", "info", "question" or "about".
    title : str, optional
        Title of messagebox.
    text : str, optional
        Text in messagebox.
    parent : QWidget, optional
        Parent widget.

    Returns
    -------
    bool
        If "OK" or "Yes" is clicked, return True. Otherwise return False.
    """
    show_dialog = _QMESSAGE_MODES[MessageBoxMode(mode)]
    result = show_dialog(parent, title, text)
    return result in (QMessageBox.Ok, QMessageBox.Yes)


def to_action(f: Callable, parent=None) -> QAction:
    action = QAction(f.__name__.replace("_", " "), parent)
    action.triggered.connect(f)
    return action


def search_name_from_qmenu(qmenu: QtW.QMenu | QtW.QMenuBar, name: str):
    for a in qmenu.children():
        if isinstance(a, QAction) and a.text() == name:
            return a
    return None


_MODIFIERS = {
    "Shift": Qt.KeyboardModifier.ShiftModifier,
    "Ctrl": Qt.KeyboardModifier.ControlModifier,
    "Control": Qt.KeyboardModifier.ControlModifier,
    "Alt": Qt.KeyboardModifier.AltModifier,
    "Meta": Qt.KeyboardModifier.MetaModifier,
}


class QtKeys:
    """A custom class for handling key events."""

    def __init__(self, e: QtGui.QKeyEvent):
        self.modifier = e.modifiers()
        self.key = e.key()

    def __eq__(self, other):
        if isinstance(other, QtKeys):
            return self.modifier == other.modifier and self.key == other.key
        elif isinstance(other, str):
            *mods, btn = other.split("+")
            if not mods:
                qtmod = Qt.KeyboardModifier.NoModifier
            else:
                qtmod = reduce(or_, [_MODIFIERS[m] for m in mods])
            qtkey = getattr(Qt.Key, f"Key_{btn}")
            return self.modifier == qtmod and self.key == qtkey
        else:
            raise TypeError

    def is_typing(self) -> bool:
        """True if key is a letter or number."""
        return (
            self.modifier
            in (
                Qt.KeyboardModifier.NoModifier,
                Qt.KeyboardModifier.ShiftModifier,
            )
            and (Qt.Key.Key_Exclam <= self.key <= Qt.Key.Key_ydiaeresis)
        )

    def key_string(self) -> str:
        """Get clicked key in string form."""
        return QtGui.QKeySequence(self.key).toString()

    def has_ctrl(self) -> bool:
        """True if Ctrl is pressed."""
        return self.modifier & Qt.KeyboardModifier.ControlModifier

    def has_shift(self) -> bool:
        """True if Shift is pressed."""
        return self.modifier & Qt.KeyboardModifier.ShiftModifier



from __future__ import annotations
from typing import TYPE_CHECKING, Callable, TypeVar
import qtpy
from qtpy import QtWidgets as QtW, QtGui, QtCore
from qtpy.QtWidgets import QAction
from qtpy.QtCore import Qt, QEvent, Signal

from ._table_stack import QTabbedTableStack
from ._utils import search_name_from_qmenu
from ..types import TabPosition

if TYPE_CHECKING:
    from ._toolbar import QTableStackToolBar
    from ..widgets import TableViewer
    from typing_extensions import ParamSpec

    _P = ParamSpec("_P")
    _R = TypeVar("_R")


class _EventFilter(QtCore.QObject):
    styleChanged = Signal()

    def eventFilter(self, obj: QtCore.QObject, event: QEvent):
        _type = event.type()
        if _type == QEvent.Type.StyleChange:
            self.styleChanged.emit()
        return False


class _QtMainWidgetBase(QtW.QWidget):
    _table_viewer: TableViewer
    _tablestack: QTabbedTableStack
    _toolbar: QTableStackToolBar

    def __init__(
        self,
        tab_position: TabPosition | str = TabPosition.top,
    ):
        super().__init__()
        self.setObjectName(f"tabulous.{type(self).__name__}")
        tab_position = TabPosition(tab_position)
        self._tablestack = QTabbedTableStack(tab_position=tab_position.name)
        self.setCentralWidget(self._tablestack)

        # NOTE: Event filter must be stored as an attribute, otherwise it will be
        # garbage collected.
        self._event_filter = _EventFilter()
        self.installEventFilter(self._event_filter)
        self._event_filter.styleChanged.connect(self.updateToolButtons)

    def updateToolButtons(self):
        if self._toolbar is None:
            return
        bg = self.palette().color(self.backgroundRole())
        whiteness = bg.red() + bg.green() + bg.blue()
        if whiteness < 128 * 3:
            self._toolbar.setToolButtonColor("#FFFFFF")
        else:
            self._toolbar.setToolButtonColor("#000000")

    def screenshot(self):
        import numpy as np

        img = self.grab().toImage()
        bits = img.constBits()
        h, w, c = img.height(), img.width(), 4
        if qtpy.API_NAME.startswith("PySide"):
            arr = np.array(bits).reshape(h, w, c)
        else:
            bits.setsize(h * w * c)
            arr = np.frombuffer(bits, np.uint8).reshape(h, w, c)

        return arr[:, :, [2, 1, 0, 3]]

    def setCentralWidget(self, wdt: QtW.QWidget):
        """Set the splitter widget."""
        raise NotImplementedError()

    def toolBarVisible(self) -> bool:
        """Visibility of toolbar"""
        raise NotImplementedError()

    def setToolBarVisible(self, visible: bool):
        """Set visibility of toolbar"""
        raise NotImplementedError()

    def addDefaultToolBar(self):
        """Add default toolbar widget regardless of self is a main window or not."""
        raise NotImplementedError()


class QMainWidget(_QtMainWidgetBase):
    def __init__(self, tab_position: TabPosition | str = TabPosition.top):
        super().__init__(tab_position)
        self._toolbar = None

    def setCentralWidget(self, wdt: QtW.QWidget):
        """Mimicking QMainWindow's method by adding a widget to the layout."""
        _layout = QtW.QVBoxLayout()
        _layout.setContentsMargins(0, 0, 0, 0)
        _layout.addWidget(wdt)
        self.setLayout(_layout)

    if TYPE_CHECKING:

        def layout(self) -> QtW.QVBoxLayout:
            ...

    def toolBarVisible(self) -> bool:
        if self._toolbar is None:
            return False
        else:
            return self._toolbar.isVisible()

    def setToolBarVisible(self, visible: bool):
        if visible and self._toolbar is None:
            from ._toolbar import QTableStackToolBar

            self._toolbar = QTableStackToolBar(self)
            self.layout().insertWidget(0, self._toolbar)

        return self._toolbar.setVisible(visible)


class QMainWindow(QtW.QMainWindow, _QtMainWidgetBase):
    _table_viewer: TableViewer
    _instances: list[QMainWindow] = []

    def __init__(
        self,
        tab_position: TabPosition | str = TabPosition.top,
    ):
        super().__init__()
        _QtMainWidgetBase.__init__(self, tab_position=tab_position)
        self.setWindowTitle("tabulous")
        from ._toolbar import QTableStackToolBar

        self._toolbar = QTableStackToolBar(self)
        self.addToolBar(self._toolbar)
        self._tablestack.setMinimumSize(400, 250)
        QMainWindow._instances.append(self)

    def addDockWidget(
        self,
        qwidget: QtW.QWidget,
        *,
        name: str = "",
        area: str = "right",
        allowed_areas: list[str] = None,
    ):
        from .._qt._dockwidget import QtDockWidget

        name = name or qwidget.objectName()
        dock = QtDockWidget(
            self,
            qwidget,
            name=name.replace("_", " "),
            area=area,
            allowed_areas=allowed_areas,
        )

        super().addDockWidget(QtDockWidget.areas[area], dock)
        return dock

    @classmethod
    def currentViewer(cls) -> TableViewer:
        """Return the current TableViewer widget."""
        window = cls._instances[-1] if cls._instances else None
        return window._table_viewer if window else None

    def event(self, e: QEvent):
        if e.type() == QEvent.Type.Close:
            # when we close the MainWindow, remove it from the instances list
            try:
                QMainWindow._instances.remove(self)
            except ValueError:
                pass
        if e.type() in {QEvent.Type.WindowActivate, QEvent.Type.ZOrderChange}:
            # upon activation or raise_, put window at the end of _instances
            try:
                inst = QMainWindow._instances
                inst.append(inst.pop(inst.index(self)))
            except ValueError:
                pass
        return super().event(e)

    def registerAction(
        self, location: str
    ) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
        locs = location.split(">")
        if len(locs) < 2:
            raise ValueError("Location must be 'XXX>YYY>ZZZ' format.")
        menu = self.menuBar()
        for loc in locs[:-1]:
            a = search_name_from_qmenu(menu, loc)
            if a is None:
                menu = menu.addMenu(loc)
            else:
                menu = a.menu()
                if menu is None:
                    i = locs.index(loc)
                    err_loc = ">".join(locs[:i])
                    raise TypeError(f"{err_loc} is not a menu.")

        def wrapper(f: Callable):
            action = QAction(locs[-1], self)
            action.triggered.connect(f)
            menu.addAction(action)
            return f

        return wrapper

    def toolBarVisible(self) -> bool:
        return self._toolbar.isVisible()

    def setToolBarVisible(self, visible: bool):
        return self._toolbar.setVisible(visible)



from __future__ import annotations
from typing import Any
import pandas as pd
from ._base import QMutableSimpleTable, DataFrameModel


class QTableLayer(QMutableSimpleTable):
    def getDataFrame(self) -> pd.DataFrame:
        return self._data_raw

    @QMutableSimpleTable._mgr.interface
    def setDataFrame(self, data: pd.DataFrame) -> None:
        self._data_raw = data
        self.model().df = data
        self.setFilter(None)
        self._qtable_view.viewport().update()
        return

    @setDataFrame.server
    def setDataFrame(self, data):
        try:
            return (self.getDataFrame(),), {}
        except Exception:
            return None

    def createModel(self):
        model = DataFrameModel(self)
        self._qtable_view.setModel(model)
        return None

    def convertValue(self, r: int, c: int, value: Any) -> Any:
        """Convert value to the type of the table."""
        kind = self._data_raw.dtypes[c].kind
        return _DTYPE_CONVERTER[kind](value)


def _bool_converter(val: Any):
    if isinstance(val, str):
        if val in ("True", "1", "true"):
            return True
        elif val in ("False", "0", "false"):
            return False
        else:
            raise ValueError(f"Cannot convert {val} to bool.")
    else:
        return bool(val)


_NAN_STRINGS = frozenset({"", "nan", "na", "n/a", "<na>"})


def _float_or_nan(x: str):
    if x.lower() in _NAN_STRINGS:
        return float("nan")
    return float(x)


def _complex_or_nan(x: str):
    if x.lower() in _NAN_STRINGS:
        return float("nan")
    return complex(x)


_DTYPE_CONVERTER = {
    "i": int,
    "f": _float_or_nan,
    "u": int,
    "b": _bool_converter,
    "U": str,
    "O": lambda e: e,
    "c": _complex_or_nan,
    "M": pd.to_datetime,
    "m": pd.to_timedelta,
}



from .mainwindow import TableViewer, TableViewerWidget
from .table import Table, SpreadSheet, GroupBy, TableDisplay

__all__ = [
    "TableViewer",
    "TableViewerWidget",
    "Table",
    "SpreadSheet",
    "GroupBy",
    "TableDisplay",
]



from __future__ import annotations
from functools import reduce
from operator import ior
from qtpy.QtWidgets import QDockWidget as _QDockWidget, QMainWindow, QWidget
from qtpy.QtCore import Qt


class QtDockWidget(_QDockWidget):
    areas = {
        "left": Qt.LeftDockWidgetArea,
        "right": Qt.RightDockWidgetArea,
        "top": Qt.TopDockWidgetArea,
        "bottom": Qt.BottomDockWidgetArea,
    }

    def __init__(
        self,
        parent: QMainWindow,
        widget: QWidget,
        *,
        name: str = "",
        area: str = "right",
        allowed_areas: list[str] = None,
    ):
        super().__init__(name, parent)
        areas = self.__class__.areas

        if allowed_areas:
            if not isinstance(allowed_areas, (list, tuple)):
                raise TypeError("`allowed_areas` must be a list or tuple")

            if any(area not in areas for area in allowed_areas):
                raise ValueError(
                    f"all allowed_areas argument must be in {set(areas.keys())}"
                )
            allowed_areas = reduce(ior, [areas[a] for a in allowed_areas])
        else:
            allowed_areas = Qt.AllDockWidgetAreas
        self.qt_area = areas[area]
        self.setAllowedAreas(allowed_areas)
        self.setWidget(widget)
        self.setMinimumHeight(50)
        self.setMinimumWidth(50)

    def setSourceObject(self, obj):
        self._widget = obj
    
    @property
    def widget(self):
        return self._widget


from __future__ import annotations
import warnings
from qtpy import QtCore, QtGui
from qtpy.QtCore import Qt, Signal
import pandas as pd

# https://ymt-lab.com/post/2020/pyqt5-qtableview-pandas-qabstractitemmodel/


class AbstractDataFrameModel(QtCore.QAbstractTableModel):
    dataEdited = Signal(int, int, object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._df = pd.DataFrame([])

    @property
    def df(self) -> pd.DataFrame:
        return self._df

    def updateValue(self, r, c, val):
        # pandas warns but no problem
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._df.iloc[r, c] = val

    def data(
        self,
        index: QtCore.QModelIndex,
        role: Qt.ItemDataRole = Qt.ItemDataRole.DisplayRole,
    ):
        if not index.isValid():
            return QtCore.QVariant()
        if role == Qt.ItemDataRole.EditRole or role == Qt.ItemDataRole.DisplayRole:
            r, c = index.row(), index.column()
            if r < self.df.shape[0] and c < self.df.shape[1]:
                val = self.df.iat[r, c]
                if pd.isna(val):
                    return "NA"
                return str(val)
            return QtCore.QVariant()
        elif role == Qt.ItemDataRole.TextColorRole:
            r, c = index.row(), index.column()
            if r < self.df.shape[0] and c < self.df.shape[1]:
                val = self.df.iat[r, c]
                if pd.isna(val):
                    return QtGui.QColor(Qt.GlobalColor.gray)
            return QtCore.QVariant()
        # elif role == Qt.ItemDataRole.ToolTipRole:
        #     ...
        return QtCore.QVariant()

    def flags(self, index):
        return (
            Qt.ItemFlag.ItemIsEditable
            | Qt.ItemFlag.ItemIsEnabled
            | Qt.ItemFlag.ItemIsSelectable
        )

    def headerData(
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ):
        if orientation == Qt.Orientation.Horizontal:
            if role == Qt.ItemDataRole.DisplayRole:
                if section < self.df.columns.size:
                    return str(self.df.columns[section])
                return None
            elif role == Qt.ItemDataRole.ToolTipRole:
                if section < self.df.columns.size:
                    name = self.df.columns[section]
                    dtype = self.df.dtypes[section]
                    return f"{name} (dtype: {dtype})"
                return None

        if orientation == Qt.Orientation.Vertical:
            if role in (Qt.ItemDataRole.DisplayRole, Qt.ItemDataRole.ToolTipRole):
                if section < self.df.index.size:
                    return str(self.df.index[section])
                return None

    def setData(self, index: QtCore.QModelIndex, value, role) -> bool:
        if not index.isValid():
            return False
        if role != Qt.ItemDataRole.EditRole:
            return False
        r, c = index.row(), index.column()
        self.dataEdited.emit(r, c, value)
        return True

    def setShape(self, nrow: int, ncol: int):
        """Set table shape."""
        r0, c0 = self.rowCount(), self.columnCount()
        dr = nrow - r0
        dc = ncol - c0

        # Adjust rows
        if dr > 0:
            self.beginInsertRows(QtCore.QModelIndex(), r0, r0 + dr - 1)
            self.insertRows(r0, dr, QtCore.QModelIndex())
            self.endInsertRows()
        elif dr < 0:
            self.beginRemoveRows(QtCore.QModelIndex(), r0 + dr, r0 - 1)
            self.removeRows(r0 + dr, -dr, QtCore.QModelIndex())
            self.endRemoveRows()

        # Adjust columns
        if dc > 0:
            self.beginInsertColumns(QtCore.QModelIndex(), c0, c0 + dc - 1)
            self.insertColumns(c0, dc, QtCore.QModelIndex())
            self.endInsertColumns()
        elif dc < 0:
            self.beginRemoveColumns(QtCore.QModelIndex(), c0 + dc, c0 - 1)
            self.removeColumns(c0 + dc, -dc, QtCore.QModelIndex())
            self.endRemoveColumns()


class DataFrameModel(AbstractDataFrameModel):
    @property
    def df(self) -> pd.DataFrame:
        return self._df

    @df.setter
    def df(self, data: pd.DataFrame):
        if data is self._df:
            return
        self.setShape(*data.shape)
        self._df = data

    def rowCount(self, parent=None):
        return self.df.shape[0]

    def columnCount(self, parent=None):
        return self.df.shape[1]



from __future__ import annotations
from typing import (
    Any,
    Callable,
    Iterable,
    NewType,
    Annotated,
    Tuple,
    List,
    Union,
    TYPE_CHECKING,
    NamedTuple,
    SupportsIndex,
)
from enum import Enum
from psygnal import Signal
from psygnal.containers import EventedList

import numpy as np
from numpy.typing import ArrayLike

if TYPE_CHECKING:
    import pandas as pd

    _TableLike = Union[pd.DataFrame, dict, Iterable, ArrayLike]
else:
    _TableLike = Any


__all__ = [
    "TableData",
    "TableColumn",
    "TableDataTuple",
    "TableInfo",
    "TablePosition",
    "ItemInfo",
    "HeaderInfo",
    "SelectionRanges",
    "SelectionType",
]

if TYPE_CHECKING:
    TableData = NewType("TableData", pd.DataFrame)
    TableColumn = NewType("TableColumn", pd.Series)
else:
    TableData = NewType("TableData", Any)
    TableColumn = NewType("TableColumn", Any)

TableDataTuple = NewType("TableDataTuple", tuple)


class TableInfoAlias(type):
    def __getitem__(cls, names: str | tuple[str, ...]):
        if isinstance(names, str):
            names = (names,)
        else:
            for name in names:
                if not isinstance(name, str):
                    raise ValueError(
                        f"Cannot subscribe type {type(name)} to TableInfo."
                    )

        return Annotated[TableInfoInstance, {"column_choice_names": names}]


class TableInfo(metaclass=TableInfoAlias):
    """
    A generic type to describe a DataFrame and its column names.

    ``TableInfo["x", "y"]`` is equivalent to ``tuple[pd.DataFrame, str, str]``
    with additional information for magicgui construction.
    """

    def __new__(cls, *args, **kwargs):
        raise TypeError(f"Type {cls.__name__} cannot be instantiated.")


class TableInfoInstance(Tuple["pd.DataFrame", List[str]]):
    def __new__(cls, *args, **kwargs):
        raise TypeError(f"Type {cls.__name__} cannot be instantiated.")


def _check_tuple_of_slices(value: Any) -> tuple[slice, slice]:
    v0, v1 = value
    if isinstance(v0, slice) and isinstance(v1, slice):
        if v0.step and v0.step != 1:
            raise ValueError("Cannot set slice with step.")
        if v1.step and v1.step != 1:
            raise ValueError("Cannot set slice with step.")
        return tuple(value)
    else:
        raise TypeError(f"Invalid input: ({type(v0)}, {type(v1)}).")


class SelectionRanges(EventedList[tuple[slice, slice]]):
    """A table data specific selection range list."""

    _changed = Signal(object)

    def __init__(self, ranges: Iterable[tuple[slice, slice]] = ()):
        super().__init__(ranges)

    def _pre_insert(self, value):
        _check_tuple_of_slices(value)
        return value

    def _post_insert(self, value) -> None:
        self._changed.emit(self)
        return value


FilterType = Union[Callable[["pd.DataFrame"], np.ndarray], np.ndarray]


class TabPosition(Enum):
    """Enum for tab position."""

    top = "top"
    left = "left"
    bottom = "bottom"
    right = "right"


class ItemInfo(NamedTuple):
    """A named tuple for item update."""

    row: int | slice
    column: int | slice
    value: Any
    old_value: Any


class HeaderInfo(NamedTuple):
    """A named tuple for header update."""

    index: int
    value: Any
    old_value: Any


_Sliceable = Union[SupportsIndex, slice]
_SingleSelection = Tuple[_Sliceable, _Sliceable]
SelectionType = List[_SingleSelection]



from __future__ import annotations
from typing import TYPE_CHECKING, Any
from io import StringIO
import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets as QtW
from qtpy.QtCore import Qt

from ._base import AbstractDataFrameModel, QMutableSimpleTable

MAX_ROW_SIZE = 100000
MAX_COLUMN_SIZE = 100000


class SpreadSheetModel(AbstractDataFrameModel):
    @property
    def df(self) -> pd.DataFrame:
        return self._df

    @df.setter
    def df(self, data: pd.DataFrame):
        self._df = data

    def rowCount(self, parent=None):
        return min(self._df.shape[0] + 10, MAX_ROW_SIZE)

    def columnCount(self, parent=None):
        return min(self._df.shape[1] + 10, MAX_COLUMN_SIZE)

    def insertRows(
        self, row: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        df = self.df
        self.beginInsertRows(parent, row, row + count - 1)
        df0 = df.iloc[:row, :]
        df1 = pd.DataFrame(
            np.full((count, df.shape[1]), np.nan), columns=df.columns, dtype="string"
        )
        df2 = df.iloc[row:, :]
        self.df = pd.concat([df0, df1, df2], axis=0)
        self.endInsertRows()
        return True

    def insertColumns(
        self, column: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        df = self.df
        self.beginInsertColumns(parent, column, column + count - 1)
        df0 = df.iloc[:, :column]
        df1 = pd.DataFrame(
            np.full((df.shape[0], count), np.nan), index=df.index, dtype="string"
        )
        df2 = df.iloc[:, column:]
        self.df = pd.concat([df0, df1, df2], axis=1)
        self.endInsertColumns()
        return True

    def removeRows(
        self, row: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        if count <= 0:
            return False
        df = self.df
        stop = row + count
        self.beginRemoveRows(parent, row, stop - 1)
        self.df = df.drop(index=df.index[row:count])
        self.endRemoveRows()
        return True

    def removeColumns(
        self, column: int, count: int, parent: QtCore.QModelIndex = None
    ) -> bool:
        if count <= 0:
            return False
        df = self.df
        stop = column + count
        self.beginRemoveColumns(parent, column, stop - 1)
        self.df = df.drop(columns=df.columns[column:stop])
        self.endRemoveColumns()
        return True


class QSpreadSheet(QMutableSimpleTable):
    """
    A table layer class that works similar to Excel sheet.

    Unlike ``QTableLayer``, this class does not have dtype. The dtype will be
    determined every time table data is converted into DataFrame. Table data
    is (almost) unbounded.
    """

    _DEFAULT_EDITABLE = True

    def __init__(self, parent=None, data: pd.DataFrame | None = None):
        super().__init__(parent, data)
        self._data_cache = None
        self._qtable_view.rightClickedSignal.connect(self.showContextMenu)

    if TYPE_CHECKING:

        def model(self) -> SpreadSheetModel:
            ...

    def getDataFrame(self) -> pd.DataFrame:
        if self._data_cache is not None:
            return self._data_cache
        # Convert table data into a DataFrame with the optimal dtypes
        buf = StringIO(self._data_raw.to_csv(sep="\t", index_label="_INDEX_"))
        out = pd.read_csv(buf, sep="\t", index_col="_INDEX_")
        out.index.name = None
        self._data_cache = out
        return out

    @QMutableSimpleTable._mgr.interface
    def setDataFrame(self, data: pd.DataFrame) -> None:
        """Set data frame as a string table."""
        self._data_raw = data.astype("string")
        self.model().setShape(data.index.size + 10, data.columns.size + 10)
        self._data_cache = None
        self.setFilter(None)
        self.refresh()
        return

    @setDataFrame.server
    def setDataFrame(self, data):
        return (getattr(self, "_data_raw", None),), {}

    def createModel(self) -> None:
        """Create spreadsheet model."""
        model = SpreadSheetModel(self)
        self._qtable_view.setModel(model)
        return None

    def convertValue(self, r: int, c: int, value: Any) -> Any:
        """Convert value to the type of the table."""
        return value

    def readClipBoard(self):
        """Read clipboard as a string data frame."""
        return pd.read_clipboard(header=None).astype("string")

    def setDataFrameValue(self, r: int | slice, c: int | slice, value: Any) -> None:
        nr, nc = self._data_raw.shape
        rmax = _get_limit(r)
        cmax = _get_limit(c)
        need_expand = nr <= rmax or nc <= cmax

        if isinstance(value, str):
            if need_expand and value == "":
                # if user start editing a cell outside the data frame and did nothing,
                # do not expand the data frame.
                return
            if isinstance(r, int) and isinstance(c, int) and value == "NA":
                # if user start editing an empty cell and did nothing, do not set string "NA".
                index = self._qtable_view.model().index(r, c, QtCore.QModelIndex())
                text = self._qtable_view.model().data(
                    index, Qt.ItemDataRole.DisplayRole
                )
                if text == value:
                    return

        with self._mgr.merging(name="setDataFrameValue"):
            if need_expand:
                self.expandDataFrame(
                    max(rmax - nr + 1, 0),
                    max(cmax - nc + 1, 0),
                )
            super().setDataFrameValue(r, c, value)
            self.setFilter(self._filter_slice)

        self._qtable_view.verticalHeader().resize(
            self._qtable_view.verticalHeader().sizeHint()
        )
        self._data_cache = None
        return None

    @QMutableSimpleTable._mgr.undoable
    def expandDataFrame(self, nrows: int, ncols: int):
        if not self._editable:
            return None
        self._data_raw = _pad_dataframe(self._data_raw, nrows, ncols)
        new_shape = self._data_raw.shape
        self.model().setShape(new_shape[0] + 10, new_shape[1] + 10)
        return None

    @expandDataFrame.undo_def
    def expandDataFrame(self, nrows: int, ncols: int):
        nr, nc = self._data_raw.shape
        self.model().removeRows(nr - nrows, nrows, QtCore.QModelIndex())
        self.model().removeColumns(nc - ncols, ncols, QtCore.QModelIndex())
        self._data_raw = self._data_raw.iloc[: nr - nrows, : nc - ncols]
        self.setFilter(self._filter_slice)
        self._data_cache = None
        return None

    def setVerticalHeaderValue(self, index: int, value: Any) -> None:
        """Set value of the table vertical header and DataFrame at the index."""
        nrows = self._data_raw.shape[0]
        if index >= nrows:
            if value == "":
                return
            self._data_raw = _pad_dataframe(self._data_raw, index - nrows + 1, 0)

        new_shape = self._data_raw.shape

        with self._mgr.blocked():
            self.setFilter(self._filter_slice)
        self.model().setShape(new_shape[0] + 10, new_shape[1] + 10)
        self._data_cache = None
        return super().setVerticalHeaderValue(index, value)

    def setHorizontalHeaderValue(self, index: int, value: Any) -> None:
        """Set value of the table horizontal header and DataFrame at the index."""
        ncols = self._data_raw.shape[1]
        if index >= ncols:
            if value == "":
                return
            self._data_raw = _pad_dataframe(self._data_raw, 0, index - ncols + 1)

        new_shape = self._data_raw.shape

        with self._mgr.blocked():
            self.setFilter(self._filter_slice)
        self.model().setShape(new_shape[0] + 10, new_shape[1] + 10)

        self._data_cache = None
        return super().setHorizontalHeaderValue(index, value)

    def showContextMenu(self, pos: QtCore.QPoint):
        menu = QtW.QMenu(self._qtable_view)
        index = self._qtable_view.indexAt(pos)
        row, col = index.row(), index.column()
        model = self.model()
        menu.addAction(
            "Insert a row above", lambda: model.insertRows(row, 1, QtCore.QModelIndex())
        )
        menu.addAction(
            "Insert a row below",
            lambda: model.insertRows(row + 1, 1, QtCore.QModelIndex()),
        )
        menu.addAction(
            "Insert a column on the left",
            lambda: model.insertColumns(col, 1, QtCore.QModelIndex()),
        )
        menu.addAction(
            "Insert a column on the right",
            lambda: model.insertColumns(col + 1, 1, QtCore.QModelIndex()),
        )
        menu.addAction(
            "Remove this row", lambda: model.removeRows(row, 1, QtCore.QModelIndex())
        )
        menu.addAction(
            "Remove this column",
            lambda: model.removeColumns(col, 1, QtCore.QModelIndex()),
        )
        return menu.exec(self._qtable_view.mapToGlobal(pos))


def _get_limit(a) -> int:
    if isinstance(a, int):
        amax = a
    elif isinstance(a, slice):
        amax = a.stop - 1
    else:
        raise TypeError(f"Cannot infer limit of type {type(a)}")
    return amax


def _pad_dataframe(df: pd.DataFrame, nr: int, nc: int) -> pd.DataFrame:
    # pad rows
    if nr > 0:
        if df.size == 0:
            df = pd.DataFrame(
                np.full((nr, 1), np.nan),
                index=range(nr),
                dtype="string",
            )
        else:
            _nr, _nc = df.shape
            ext = pd.DataFrame(
                np.full((nr, _nc), np.nan),
                index=range(_nr, _nr + nr),
                columns=df.columns,
                dtype="string",
            )
            df = pd.concat([df, ext], axis=0)

    # pad columns
    if nc > 0:
        if df.size == 0:
            df = pd.DataFrame(
                np.full((1, nc), np.nan),
                columns=range(nc),
                dtype="string",
            )
        else:
            _nr, _nc = df.shape
            ext = pd.DataFrame(
                np.full((_nr, nc), np.nan),
                index=df.index,
                columns=range(_nc, _nc + nc),
                dtype="string",
            )
            df = pd.concat([df, ext], axis=1)

    return df



from __future__ import annotations
from qtpy.QtCore import Qt
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QShortcut
from enum import Enum
from typing import Callable, NewType, Union, Tuple, TYPE_CHECKING
import re

if TYPE_CHECKING:
    from qtpy.QtWidgets import QWidget

QtKey = NewType("QtKey", int)
QtModifier = NewType("QtModifier", int)
KeyCombo = Union[
    Tuple[QtKey], Tuple[QtModifier, QtKey], Tuple[QtModifier, QtModifier, QtKey]
]


class Key(Enum):
    # keys
    Tab = "tab"
    Backspace = "backspace"
    Delete = "delete"
    Left = "left"
    Up = "up"
    Right = "right"
    Down = "down"
    PageUp = "pageup"
    PageDown = "pagedown"
    F1 = "f1"
    F2 = "f2"
    F3 = "f3"
    F4 = "f4"
    F5 = "f5"
    F6 = "f6"
    F7 = "f7"
    F8 = "f8"
    F9 = "f9"
    F10 = "f10"
    F11 = "f11"
    F12 = "f12"
    Exclam = "!"
    Dollar = "$"
    Percent = "%"
    Ampersand = "&"
    Apostrophe = "'"
    ParenLeft = "("
    ParenRight = ")"
    Asterisk = "*"
    Plus = "+"
    Comma = ","
    Minus = "-"
    Period = "."
    Slash = "/"
    Colon = ":"
    Semicolon = ";"
    Less = "<"
    Equal = "="
    Greater = ">"
    Question = "?"
    At = "@"

    # alphabets
    A = "a"
    B = "b"
    C = "c"
    D = "d"
    E = "e"
    F = "f"
    G = "g"
    H = "h"
    I = "i"
    J = "j"
    K = "k"
    L = "l"
    M = "m"
    N = "n"
    O = "o"
    P = "p"
    Q = "q"
    R = "r"
    S = "s"
    T = "t"
    U = "u"
    V = "v"
    W = "w"
    X = "x"
    Y = "y"
    Z = "z"

    # numbers
    _0 = "0"
    _1 = "1"
    _2 = "2"
    _3 = "3"
    _4 = "4"
    _5 = "5"
    _6 = "6"
    _7 = "7"
    _8 = "8"
    _9 = "9"

    # modifiers
    Meta = "meta"
    Shift = "shift"
    Ctrl = "ctrl"
    Alt = "alt"

    @classmethod
    def to_qtkey(cls, key: str | int | Key) -> QtKey:
        if isinstance(key, str):
            key = cls(key.lower())
        elif isinstance(key, int):
            key = str(key)
        elif isinstance(key, cls):
            pass
        else:
            raise TypeError(f"Unsupported type for a key: {type(key)}.")
        return getattr(Qt, f"Key_{key.name.lstrip('_')}")

    @classmethod
    def to_qtmodifier(cls, key: str | Key) -> QtModifier:
        if key == "control":
            key = cls.Ctrl
        elif isinstance(key, str):
            key = cls(key.lower())
        elif isinstance(key, cls):
            pass
        else:
            raise TypeError(f"Unsupported type for a modifier: {type(key)}.")
        return getattr(Qt, key.name.upper())

    # Add method enables expressions like ``Key.Ctrl + Key.A``.
    def __add__(self, other: str | Key) -> tuple[Key, Key]:
        cls = self.__class__
        if isinstance(other, str):
            other = cls(other.lower())
        elif not isinstance(other, cls):
            raise TypeError(f"Cannot add type {type(other)} to Key object.")
        return (self, other)

    def __radd__(self, other: str | Key | tuple[str | Key, ...]) -> tuple[Key, ...]:
        cls = self.__class__
        if isinstance(other, str):
            other = cls(other.lower())
            return (self, other)
        elif isinstance(other, cls):
            return (other, self)
        elif isinstance(other, tuple):
            return other + (self,)
        else:
            raise TypeError(f"Cannot add Key object to type {type(other)}.")


MODIFIERS = (Key.Meta, Key.Shift, Key.Ctrl, Key.Alt)


def ismodifier(s: str) -> bool:
    s = s.lower()
    if s == "control":
        return True
    else:
        return Key(s) in MODIFIERS


def parse_key_combo(key_combo: str) -> QtKey:
    # For compatibility with napari
    parsed = re.split("-(?=.+)", key_combo)
    return strs2keycombo(*parsed)


def strs2keycombo(*args: tuple[str | Key, ...]) -> KeyCombo:
    *modifiers, key = args
    if len(modifiers) > 2:
        raise ValueError("More than two modifiers found.")
    return tuple(Key.to_qtmodifier(m) for m in modifiers) + (Key.to_qtkey(key),)


def as_key_sequence(key_combo: tuple) -> QKeySequence:
    if not isinstance(key_combo, tuple):
        raise TypeError(f"Unsupported key combo: {key_combo!r}.")

    key0 = key_combo[0]
    if len(key_combo) == 1 and isinstance(key0, str) and not hasattr(Key, key0.lower()):
        qtkeycombo = parse_key_combo(*key_combo)
    else:
        qtkeycombo = strs2keycombo(*key_combo)
    return QKeySequence(sum(qtkeycombo))


def register_shortcut(keys, parent: QWidget, target: Callable):
    """Register a callback to a key-binding globally."""
    shortcut = QShortcut(as_key_sequence(keys), parent)
    shortcut.activated.connect(target)
    return None



from qtpy.QtWidgets import QApplication
from qtpy.QtCore import Qt
from qtpy import PYQT5

APPLICATION = None


def gui_qt():
    """Call "%gui qt" magic."""
    try:
        from IPython import get_ipython
    except ImportError:
        get_ipython = lambda: False

    shell = get_ipython()

    if shell and shell.active_eventloop != "qt":
        shell.enable_gui("qt")
    return None


def gui_qt_is_active() -> bool:
    """True only if "%gui qt" magic is called in ipython kernel."""
    try:
        from IPython import get_ipython
    except ImportError:
        get_ipython = lambda: False

    shell = get_ipython()

    return shell and shell.active_eventloop == "qt"


def get_app():
    """Get QApplication."""
    global APPLICATION
    gui_qt()
    app = QApplication.instance()
    if app is None:
        if PYQT5:
            QApplication.setAttribute(Qt.AA_EnableHighDpiScaling)
            QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps)
        app = QApplication([])
    APPLICATION = app
    return app


def run_app():
    """Start the event loop."""
    if not gui_qt_is_active():
        return get_app().exec_()



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable
import operator as op

from ..types import FilterType

if TYPE_CHECKING:
    from .table import TableBase
    import pandas as pd


class FilterProxy:
    """
    A proxy of column-based DataFrame filtering.

    This object makes filtering of table easier. For instance, to filter a table
    by a column, you can do:
    >>> table.filter["a"] < 5
    which is equivalent to
    >>> table.filter = lambda df: df["a"] < 5
    """

    def __init__(self, table: TableBase | None = None):
        self._table = table

    def __repr__(self) -> str:
        return f"{type(self).__name__}(table={self._table!r}, func={self.func!r})"

    def __bool__(self) -> bool:
        return self.func is not None

    def __get__(self, obj: TableBase | None, type=None):
        return self.__class__(obj)

    def __set__(self, obj: TableBase | None, value: FilterType):
        if obj is None:
            return None
        data = obj.data
        if callable(value):
            # dry run
            try:
                df = data.head(3)
                filt = value(df)
            except Exception as e:
                raise ValueError(
                    f"Dry run failed with filter function {value} due to following error:\n"
                    f"{type(e).__name__}: {e}"
                ) from None

        elif isinstance(value, FilterProxy):
            value = value.func
        elif value is not None and len(value) != data.shape[0]:
            raise ValueError(
                f"Shape mismatch between data {data.shape} and input slice {len(value)}."
            )
        obj._qwidget.setFilter(value)

    def __delete__(self, obj: TableBase):
        """Disable filter."""
        obj.filter = None

    def __getitem__(self, key):
        return FilterIndexer(self._table, key)

    @property
    def func(self) -> FilterType:
        """Get the filter function."""
        return self._table._qwidget.filter()

    @property
    def table(self) -> TableBase | None:
        """Return the connected table."""
        return self._table

    def array(self) -> pd.Series:
        f = self.func
        if callable(f):
            arr = f(self._table.data)
        else:
            arr = f
        return arr


class FilterIndexer:
    def __init__(
        self, table: TableBase, key: Any, *, current_filter: ColumnFilter | None = None
    ):
        self._table = table
        self._key = key
        self._filter = current_filter
        self._table._qwidget.setFilter(current_filter)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} of {self._table!r} at column {self._key!r}>"

    def chain(self, other: ColumnFilter) -> ColumnFilter:
        if self._filter is None:
            fil = other
        else:
            fil = self._filter & other
        return self.filter(fil)

    def filter(self, filter: ColumnFilter) -> FilterIndexer:
        return FilterIndexer(self._table, self._key, current_filter=filter)

    def __eq__(self, other) -> ColumnFilter:
        fil = ColumnFilter.from_operator("__eq__", self._key, other)
        return self.chain(fil)

    def __gt__(self, other) -> ColumnFilter:
        fil = ColumnFilter.from_operator("__gt__", self._key, other)
        return self.chain(fil)

    def __ge__(self, other) -> ColumnFilter:
        fil = ColumnFilter.from_operator("__ge__", self._key, other)
        return self.chain(fil)

    def __lt__(self, other) -> ColumnFilter:
        fil = ColumnFilter.from_operator("__lt__", self._key, other)
        return self.chain(fil)

    def __le__(self, other) -> ColumnFilter:
        fil = ColumnFilter.from_operator("__le__", self._key, other)
        return self.chain(fil)

    def __ne__(self, other) -> ColumnFilter:
        fil = ColumnFilter.from_operator("__ne__", self._key, other)
        return self.chain(fil)

    def __and__(self, other: FilterIndexer) -> FilterIndexer:
        fil = self._filter & other._filter
        return self.filter(fil)

    def __or__(self, other: FilterIndexer) -> FilterIndexer:
        fil = self._filter | other._filter
        return self.filter(fil)

    def __contains__(self, other) -> None:
        fil = ColumnFilter(
            lambda df: df[self._key].isin(other),
            repr=f"df[{self._key}].isin({list(other)!r})",
        )
        return self.chain(fil)

    # TODO: __invert__, __mod__


class ColumnFilter:
    """A callable object compatible with DataFrame filter, specific to column based filter."""

    def __init__(
        self,
        func: Callable[[pd.DataFrame], pd.Series],
        repr: str | None = None,
    ):
        self._func = func
        self._repr = repr

    def __call__(self, df: pd.DataFrame) -> pd.Series:
        series = self._func(df)
        if self._repr is not None:
            series.name = self._repr
        return series

    def __repr__(self) -> str:
        return f"<{type(self).__name__} of {self._repr}>"

    def __and__(self, other: ColumnFilter) -> ColumnFilter:
        return ColumnFilter(
            lambda df: self(df) & other(df),
            repr=f"{self._repr} & {other._repr}",
        )

    def __or__(self, other: ColumnFilter) -> ColumnFilter:
        return ColumnFilter(
            lambda df: self(df) | other(df),
            repr=f"{self._repr} | {other._repr}",
        )

    @classmethod
    def from_operator(self, operator_name: str, key: Any, other: Any) -> ColumnFilter:
        """
        Construct an instance from a operator.

        Parameters
        ----------
        operator_name : str
            Name of operator, such as "__eq__".
        key : Any
            Target column name.
        other : Any
            The other argument of the operator.

        Returns
        -------
        ColumnFilter
            An instance initialized with the given operator.
        """
        f = getattr(op, operator_name)
        _op_ = OPERATORS[operator_name]
        return ColumnFilter(
            lambda df: f(df[key], other),
            repr=f"df[{key!r}] {_op_} {other!r}",
        )


OPERATORS = {
    "__eq__": "==",
    "__gt__": ">",
    "__ge__": ">=",
    "__lt__": "<",
    "__le__": "<=",
    "__ne__": "!=",
}



from __future__ import annotations
from qtpy import QtWidgets as QtW, QtCore


def create_temporal_line_edit(
    rect: QtCore.QRect,
    parent: QtW.QWidget, 
    text: str,
) -> QtW.QLineEdit:
    line = QtW.QLineEdit(parent=parent)
    geometry = line.geometry()
    geometry.setWidth(rect.width())
    geometry.setHeight(rect.height())
    geometry.moveCenter(rect.center())
    # edit_geometry.moveTo(rect.topLeft())
    line.setGeometry(geometry)
    line.setText(text)
    line.setHidden(False)
    line.setFocus()
    line.selectAll()
    return line



from __future__ import annotations
from functools import partial
from pathlib import Path
import weakref
from enum import Enum
from typing import TYPE_CHECKING, Any, Callable, Union
from psygnal import Signal, SignalGroup

from .table import Table, SpreadSheet, GroupBy, TableDisplay
from .tablelist import TableList
from .keybindings import register_shortcut
from ._sample import open_sample

from ..types import SelectionType, TabPosition, _TableLike, _SingleSelection

if TYPE_CHECKING:
    from .table import TableBase
    from .._qt import QMainWindow, QMainWidget
    from .._qt._dockwidget import QtDockWidget
    from .._qt._mainwindow import _QtMainWidgetBase
    from qtpy.QtWidgets import QWidget
    from magicgui.widgets import Widget
    import numpy as np
    import pandas as pd

PathLike = Union[str, Path, bytes]


class TableType(Enum):
    table = "table"
    spreadsheet = "spreadsheet"


class TableViewerSignal(SignalGroup):
    """Signal group for table viewer."""

    current_index = Signal(int)


class Toolbar:
    """The toolbar API."""

    def __init__(self, parent: _TableViewerBase):
        self.parent = parent

    def __repr__(self) -> str:
        return f"<{type(self).__name__} of {self.parent!r}>"

    @property
    def visible(self) -> bool:
        return self.parent._qwidget.toolBarVisible()

    @visible.setter
    def visible(self, val) -> None:
        return self.parent._qwidget.setToolBarVisible(val)


class _TableViewerBase:
    events: TableViewerSignal
    _qwidget_class: type[_QtMainWidgetBase]

    def __init__(
        self, *, tab_position: TabPosition | str = TabPosition.top, show: bool = True
    ):
        from .._qt import get_app

        app = get_app()
        self._qwidget = self._qwidget_class(tab_position=tab_position)
        self._qwidget._table_viewer = self
        self._tablist = TableList(parent=self)
        self._toolbar = Toolbar(parent=self)
        self._link_events()

        self.events = TableViewerSignal()

        if show:
            self.show(run=False)

    def __repr__(self) -> str:
        return f"<{type(self).__name__} widget at {hex(id(self))}>"

    def reset_choices(self, *_):
        pass

    @property
    def tables(self) -> TableList:
        """Return the table list object."""
        return self._tablist

    @property
    def toolbar(self) -> Toolbar:
        """Return the tool bar widget."""
        return self._toolbar

    @property
    def current_table(self) -> TableBase:
        """Return the currently visible table."""
        return self.tables[self.current_index]

    @property
    def current_index(self) -> int:
        """Return the index of currently visible table."""
        return self._qwidget._tablestack.currentIndex()

    @property
    def native(self) -> _QtMainWidgetBase:
        """Return the native widget."""
        return self._qwidget

    @current_index.setter
    def current_index(self, index: int | str):
        if isinstance(index, str):
            index = self.tables.index(index)
        elif index < 0:
            index += len(self.tables)
        return self._qwidget._tablestack.setCurrentIndex(index)

    def bind_key(self, *seq) -> Callable[[TableViewer], Any | None]:
        # TODO
        def register(f):
            register_shortcut(seq, self._qwidget, partial(f, self))

        return register

    def show(self, *, run: bool = True) -> None:
        """Show the widget."""
        self._qwidget.show()
        if run:
            from .._qt._app import run_app

            run_app()
        return None

    def screenshot(self) -> np.ndarray:
        """Get screenshot of the widget."""
        return self._qwidget.screenshot()

    def add_table(
        self,
        data: _TableLike | None = None,
        *,
        name: str | None = None,
        editable: bool = False,
        copy: bool = True,
    ) -> TableBase:

        """
        Add data as a table.

        Parameters
        ----------
        data : DataFrame like, optional
            Table data to add.
        name : str, optional
            Name of the table.
        editable : bool, default is False
            Whether the table is editable via UI.
        copy : bool, default is True
            Whether to copy the data before adding to avoid overwriting the original one.

        Returns
        -------
        TableLayerBase
            The added table object.
        """
        if copy:
            data = _copy_dataframe(data)
        table = Table(data, name=name, editable=editable)
        return self.add_layer(table)

    def add_spreadsheet(
        self,
        data: _TableLike | None = None,
        *,
        name: str | None = None,
        editable: bool = True,
        copy: bool = True,
    ) -> SpreadSheet:
        """
        Add data as a spreadsheet.

        Parameters
        ----------
        data : DataFrame like, optional
            Table data to add.
        name : str, optional
            Name of the table.
        editable : bool, default is False
            Whether the table is editable via UI.
        copy : bool, default is True
            Whether to copy the data before adding to avoid overwriting the original one.

        Returns
        -------
        SpreadSheet
            The added table object.
        """
        if copy:
            data = _copy_dataframe(data)
        table = SpreadSheet(data, name=name, editable=editable)
        return self.add_layer(table)

    def add_groupby(self, data, name: str | None = None) -> GroupBy:
        table = GroupBy(data, name=name)
        return self.add_layer(table)

    def add_loader(self, loader, name: str | None = None) -> TableDisplay:
        table = TableDisplay(loader, name=name)
        return self.add_layer(table)

    def add_layer(self, layer: TableBase):
        self.tables.append(layer)
        self.current_index = -1  # activate the last table
        return layer

    def open(self, path: PathLike, *, type: TableType | str = TableType.table) -> None:
        """
        Read a table data and add to the viewer.

        Parameters
        ----------
        path : path like
            File path.
        """
        path = Path(path)
        suf = path.suffix
        type = TableType(type)
        if type == TableType.table:
            fopen = self.add_table
        elif type == TableType.spreadsheet:
            fopen = self.add_spreadsheet
        else:
            raise RuntimeError

        import pandas as pd

        if suf in (".csv", ".txt", ".dat"):
            df = pd.read_csv(path)
            fopen(df, name=path.stem)
        elif suf in (".xlsx", ".xls", ".xlsb", ".xlsm", ".xltm", "xltx", ".xml"):
            df_dict: dict[str, pd.DataFrame] = pd.read_excel(path, sheet_name=None)
            for sheet_name, df in df_dict.items():
                fopen(df, name=sheet_name)
        else:
            raise ValueError(f"Extension {suf} not supported.")

    def save(self, path: PathLike) -> None:
        """Save current table."""
        path = Path(path)
        suf = path.suffix
        df = self.current_table.data
        if suf in (".csv", ".txt", ".dat"):
            df.to_csv(path)
        elif suf in (".xlsx", ".xls", "xml"):
            df.to_excel(path)
        else:
            raise ValueError(f"Extension {suf} not supported.")

    def open_sample(self, sample_name: str, plugin: str = "seaborn") -> Table:
        """Open a sample table."""
        df = open_sample(sample_name, plugin)
        return self.add_table(df, name=sample_name)

    def copy_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
        *,
        headers: bool = False,
    ) -> None:
        """Copy selected cells to clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.copyToClipboard(headers=headers)

    def paste_data(
        self,
        selections: SelectionType | _SingleSelection | None = None,
    ) -> None:
        """Paste from clipboard."""
        if selections is not None:
            self.current_table.selections = selections
        return self.current_table._qwidget.pasteFromClipBoard()

    def _link_events(self):
        _tablist = self._tablist
        _qtablist = self._qwidget._tablestack

        @_tablist.events.inserted.connect
        def _insert_qtable(i: int):
            table = _tablist[i]
            _qtablist.addTable(table._qwidget, table.name)

        @_tablist.events.removed.connect
        def _remove_qtable(index: int, table: TableBase):
            with _tablist.events.blocked():
                _qtablist.takeTable(index)

        @_tablist.events.moved.connect
        def _move_qtable(src: int, dst: int):
            with _tablist.events.blocked():
                _qtablist.moveTable(src, dst)

        @_tablist.events.renamed.connect
        def _rename_qtable(index: int, name: str):
            with _tablist.events.blocked():
                _qtablist.renameTable(index, name)

        @_qtablist.itemMoved.connect
        def _move_pytable(src: int, dst: int):
            """Move evented list when list is moved in GUI."""
            with self._tablist.events.blocked():
                if dst > src:
                    dst += 1
                self._tablist.move(src, dst)

        @_qtablist.tableRenamed.connect
        def _rename_pytable(index: int, name: str):
            self._tablist.rename(index, name)

        @_qtablist.tableRemoved.connect
        def _remove_pytable(index: int):
            with self._tablist.events.blocked():
                del self._tablist[index]

        @_qtablist.tablePassed.connect
        def _pass_pytable(src, index: int, dst):
            src_ = _find_parent_table(src)
            dst_ = _find_parent_table(dst)
            dst_.tables.append(src_.tables.pop(index))

        _qtablist.itemDropped.connect(self.open)

        # reset choices when something changed in python table list
        _tablist.events.inserted.connect(self.reset_choices)
        _tablist.events.removed.connect(self.reset_choices)
        _tablist.events.moved.connect(self.reset_choices)
        _tablist.events.changed.connect(self.reset_choices)
        _tablist.events.renamed.connect(self.reset_choices)


class TableViewerWidget(_TableViewerBase):
    """The non-main table viewer widget."""

    events: TableViewerSignal
    _qwidget: QMainWidget

    @property
    def _qwidget_class(self) -> QMainWidget:
        from .._qt import QMainWidget

        return QMainWidget

    def add_widget(
        self,
        widget: Widget | QWidget,
        *,
        name: str = "",
    ):
        backend_widget, name = _normalize_widget(widget, name)
        backend_widget.setParent(self._qwidget, backend_widget.windowFlags())
        return backend_widget


class TableViewer(_TableViewerBase):
    """The main table viewer widget."""

    events: TableViewerSignal
    _dock_widgets: weakref.WeakValueDictionary[str, QtDockWidget]
    _qwidget: QMainWindow

    @property
    def _qwidget_class(self) -> QMainWindow:
        from .._qt import QMainWindow

        return QMainWindow

    def __init__(
        self,
        *,
        tab_position: TabPosition | str = TabPosition.top,
        show: bool = True,
    ):
        self._dock_widgets = weakref.WeakValueDictionary()
        super().__init__(
            tab_position=tab_position,
            show=show,
        )

    # def register_action(self, location: str):
    #     return self._qwidget.registerAction(location)

    def add_dock_widget(
        self,
        widget: Widget | QWidget,
        *,
        name: str = "",
        area: str = "right",
        allowed_areas: list[str] = None,
    ):
        backend_widget, name = _normalize_widget(widget, name)

        dock = self._qwidget.addDockWidget(
            backend_widget, name=name, area=area, allowed_areas=allowed_areas
        )
        dock.setSourceObject(widget)
        self._dock_widgets[name] = dock
        return dock

    def remove_dock_widget(self, name_or_widget):
        if isinstance(name_or_widget, str):
            name = name_or_widget
            dock = self._dock_widgets[name_or_widget]
        else:
            for k, v in self._dock_widgets.items():
                if v is name_or_widget:
                    name = k
                    dock = v
                    break
            else:
                raise ValueError(f"Widget {name_or_widget} not found.")
        self._qwidget.removeDockWidget(dock)
        self._dock_widgets.pop(name)
        return None

    def reset_choices(self, *_):
        for dock in self._dock_widgets.values():
            widget = dock.widget
            if hasattr(widget, "reset_choices"):
                widget.reset_choices()


def _normalize_widget(widget: Widget | QWidget, name: str) -> tuple[QWidget, str]:
    if hasattr(widget, "native"):
        backend_widget = widget.native
        if not name:
            name = widget.name
    else:
        backend_widget = widget
        if not name:
            name = backend_widget.objectName()

    return backend_widget, name


def _copy_dataframe(data) -> pd.DataFrame:
    import pandas as pd

    return pd.DataFrame(data)


def _find_parent_table(qwidget: _QtMainWidgetBase) -> _TableViewerBase:
    x = qwidget
    while (parent := x.parent()) is not None:
        x = parent
        if hasattr(x, "_table_viewer"):
            return x._table_viewer
    raise RuntimeError
